import pytest

@pytest.fixture(scope="session", autouse=True)
def set_up_overall(request):
    request.addfinalizer(tear_down)

def tear_down():
    import os
    import glob

    patterns = ['*pdb', '*csv', '*.yaml']
    pytest_scratches = []
    for pat in patterns:
        pytest_scratches.extend(glob.glob(pat))
    for fl in pytest_scratches:
        os.unlink(fl)



"""!@file
This is the main file for the pnab driver

@namespace driver
@brief This is the main file for the pnab driver

This file contains the pNAB class for calling the C++ library. 
"""

from __future__ import division, absolute_import, print_function

import os
import yaml
import itertools
from io import StringIO
import multiprocessing as mp
import warnings
warnings.filterwarnings("ignore", category=UserWarning)
import datetime
import time
import numpy as np
import copy

from pnab import __path__
from pnab import bind
from pnab.driver import options

# Catch interruption; does not work properly for windows
import signal
def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

class pNAB(object):
    """!@brief The proto-Nucleic Acid Builder main python class

    A class that contains methods to create a pNAB run. It validates input,
    runs it, and save results.

    Examples:
    Use the code to get RNA and DNA candidates. This uses the provided example files
    in the "pnab/data" directory. The results are in three files:

    1) results.csv: Contains all the results for all the helical configurations
        requested in the run.

    2) prefix.yaml: Containts a dictionary of the sequence of the run and the 
        corresponding helical configuration (useful when a range of values is
        given for the helical parameters).

    @code
    import pnab

    run = pnab.pNAB('RNA.yaml')
    run.run()
    @endcode

    @code
    import pnab

    run = pnab.pNAB('DNA.yaml')
    run.run()
    @endcode
    """

    def __init__(self, file_path):
        """!@brief The constructor for the pNAB run

        The constructor takes an input file as an argument and initialize a new instance.
        The input file uses a YAML format for specifying the run options. This input
        file can be generated using the graphical user interface created in @a jupyter_widgets.
        Additionally, there are a few example files in the "pnab/data" directory.
        The constructor creates the @a pNAB.pNAB.options attribute which contains a dictionary of all
        the defined options.

        @param file_path (str) Path to a file containing the user-defined input. This function
            tries to find the file in the current working directory. If it does not exist,
            it looks for the file in the "pnab/data" directory

        @sa jupyter_widgets
        @sa options.validate_all_options
        """

        if not os.path.isfile(file_path):
            file_path = os.path.join(__path__[0], 'data', file_path)

        ##@brief Validated options dictionary for the pNAB run
        #@sa options.validate_all_options
        self.options = yaml.load(open(file_path, 'r'), yaml.FullLoader)

        ##@brief A string comma-separated header for the results used in the generated CSV files
        self.header = ('Prefix, Conformer Index, Distance (Angstroms), Bond Energy (kcal/mol), Angle Energy (kcal/mol), ' +
                       'Torsion Energy (kcal/mol/nucleotide), Van der Waals Energy (kcal/mol/nucleotide), ' + 
                       'Total Energy (kcal/mol/nucleotide)')

    def _run(self, config):
        """!@brief Function to run one helical configuration.

        Setup the options to call the C++ code using the pybind11 bind class in @a binder.cpp.
        After the run finishes, it writes the results.
        Two files are written: 
            "prefix.yaml" contains a dictionary of the sequence of the run and the corresponding helical configuration
            "results.csv" contains the information on the accepted candidates for all of configurations

        @param config (list) a list containing two entries: the list of helical parameter values for this
            configuration, and a string index for the sequnce of the run

        @sa run
        """
        config, prefix = config[0], config[1]

        # Add a header of the helical parameters
        header_dict = {'%s' %k:  '%.2f' %val for k, val in zip(self._options['HelicalParameters'], config)}
        units = ["A", "A", "A", "deg", "deg", "deg"]
        if self._is_helical:
            header = ''.join(['%s (%s): %s; ' %(k, units[i], header_dict[k]) for i, k in enumerate(['x_displacement', 'y_displacement', 'h_rise', 'inclination', 'tip', 'h_twist'])])
        else:
            header = ''.join(['%s (%s): %s; ' %(k, units[i], header_dict[k]) for i, k in enumerate(['shift', 'slide', 'rise', 'tilt', 'roll', 'twist'])])
        header += ''.join(['%s (%s): %s; ' %(k, units[i], header_dict[k]) for i, k in enumerate(['shear', 'stretch', 'stagger', 'buckle', 'propeller', 'opening'])])
        header = header.strip('; ')

        # Set runtime parameters
        runtime_parameters = bind.RuntimeParameters()
        [runtime_parameters.__setattr__(k, val) for k, val in self._options['RuntimeParameters'].items()]

        # Set backbone parameters
        backbone = bind.Backbone()
        [backbone.__setattr__(k, val) for k, val in self._options['Backbone'].items()]

        # Set a list of all defined bases asd there parameters
        # We include all the bases even though not all of them are requested by the user
        py_bases = [self._options[i] for i in self._options if 'Base' in i]
        bases = [bind.Base() for i in range(len(py_bases))]
        for i, b in enumerate(py_bases):
            [bases[i].__setattr__(k, val) for k, val in b.items()]

        # Set helical parameters
        helical_parameters = bind.HelicalParameters()
        [helical_parameters.__setattr__(k, val) for k, val in zip(self._options['HelicalParameters'], config)]
        helical_parameters.is_helical = self._is_helical

        # Run code
        
        result = bind.run(runtime_parameters, backbone, bases, helical_parameters, prefix, self._verbose)

        # Get results from the comma-separated output string
        result = np.genfromtxt(StringIO(result), delimiter=',')

        if result.size == 0:
            result = None
        elif result.ndim == 1:
            result = result.reshape(1, len(result))

        if result is not None:
            result = result[result[:, 7].argsort()]

        results = [prefix, header, result]

        # Write prefix and helical configuration
        with open('prefix.yaml', 'ab') as f:
            f.write(str.encode(yaml.dump({results[0]:results[1]})))

        # If there are no results, return
        if results[2] is None:
            return

        # Update header
        header = results[1] + '\n'

        header2 = self.header
        ## Add header entries for the dihedral angles
        for i in range(results[2].shape[1] - 8):
            header2 += ", Dihedral " + str(i+1) + " (degrees)"

        header += header2

        # Write results to file
        fmt = "%i,%i"
        for i in range(results[2].shape[1] - 2):
            fmt += ",%.6f"
        with open('results.csv', 'ab') as f:
            np.savetxt(f, results[2], header=header, fmt=fmt)

    def run(self, number_of_cpus=None, verbose=True, interrupt=False):
        """!@brief Prepare helical configurations and run them in parallel.

        This function first copies the user-defined options to an internal options dictionary (self._options).
        Next, self._options is validated through a call to @a options.validate_all_options to
        validate all the options for the backbone parameters, helical parameters, and runtime
        parameters. Then, it adds the library of all the defined nucleobases.
        The nucleobases are defined in the "pnab/data" directory. Whether to pair
        adenine with thymine (default) or uracil is determined here by the @a 
        pNAB.pNAB.options['RuntimeParameters']['pair_A_U'].

        If a single value is given for a helical parameter (e.g. helical twist),
        then that value is used. If a range of values is given, then equally spaced values
        in the range will be used. The number of configurations is determined by the
        third value in @a pNAB.pNAB.options. The various helical configurations
        are run in parallel using the multiprocessing library. 

        This function writes three output files: "results.csv" and "prefix.yaml".
        It renames any existing files with these names by prepending enough "_".

        @param number_of_cpus Number of CPUs to use for parallel computations of different helical configurations, defaults to all cores
        @param verbose Whether to print progress report to the screen, default to True
        @param interrupt How to handle keyboard interrupt in multiprocessing

        @returns None; output files are written
        """

        self._options = copy.deepcopy(self.options)
        options.validate_all_options(self._options)

        # Add library of bases
        data_dir = os.path.join(__path__[0], 'data')
        bases_lib = yaml.load(open(os.path.join(data_dir, 'bases_library.yaml'), 'r'), yaml.FullLoader)
        for b in bases_lib.values():
            b['file_path'] = os.path.join(data_dir, b['file_path'])
        if self._options['RuntimeParameters'].pop('pair_A_U', None):
            bases_lib['Base A']['pair_name'] = 'U'
        self._options.update(bases_lib)

        ##@brief Whether to print progress report to the screen
        self._verbose = verbose

        self._is_helical = self._options['HelicalParameters'].pop('is_helical')
        hp = self._options['HelicalParameters'].copy()
        if self._is_helical:
            # Set the number of configurations for each of the step parameters to 1
            # so that we don't generate more configurations than necessary in case
            # the user specifies both in the input file
            for k in ['shift', 'slide', 'rise', 'tilt', 'roll', 'twist']:
                hp[k][2] = 1
        else:
            for k in ['x_displacement', 'y_displacement', 'h_rise', 'inclination', 'tip', 'h_twist']:
                hp[k][2] = 1

        # Extract configurations
        config = itertools.product(*[np.linspace(val[0], val[1], val[2]) for val in hp.values()])
        num_config = np.prod([val[2] for val in hp.values()])
        prefix = (str(i) for i in range(1, num_config + 1))

        number_of_cpus = mp.cpu_count() if number_of_cpus is None else number_of_cpus

        # I cannot handle keyboard interrupt in jupyter notebook and still get results
        # I need init_worker function to gracefully terminate Jupyter notebook run
        # For command line, I can extract the results even when ctrl+c is applied,
        # in which case interrupt variable is false
        # Use maxtasksperchild=1 to free memory after each run
        if interrupt:
            pool = mp.Pool(number_of_cpus, init_worker, maxtasksperchild=1)
        else:
            pool = mp.Pool(number_of_cpus, maxtasksperchild=1)

        # Rename output files that have the same name
        for f in ['results.csv', 'prefix.yaml']:
            file_path = f
            while True:
                if os.path.isfile(file_path):
                    file_path = '_' + file_path
                else:
                    if os.path.isfile(f):
                        os.rename(f, file_path)
                    break

        # Write time stamps
        current_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        with open('results.csv', 'w') as f: f.write('# ' + current_time + '\n')
        with open('prefix.yaml', 'w') as f: f.write('# ' + current_time + '\n')

        try:
            # Run the different helical configurations in parallel and process results
            for results in pool.imap_unordered(self._run, zip(config, prefix)):
                pass

        except KeyboardInterrupt:
            # If interuption is catched, terminate run and proceed
            print("Caught interruption; stopping ...")
            time.sleep(1) # Sleep one second to allow all running processes to finish
            pool.terminate()

        except RuntimeError as e:
            # Error raised in the C++ code
            raise RuntimeError(e)

        pool.close()
        pool.join()

        # Write time stamps
        current_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        with open('results.csv', 'a') as f: f.write('# ' + current_time + '\n')
        with open('prefix.yaml', 'a') as f: f.write('# ' + current_time + '\n')

        del self._is_helical

        #Extract the results from the run

        ##@brief A dictionary of the sequence of the run and the 
        # corresponding helical configuration
        # For runs that have ranges of values for the helical parameters,
        # this dictionary provides a mapping between the sequence of the run
        # and the helical parameters.
        self.prefix = yaml.load(open('prefix.yaml'), yaml.FullLoader)

        ##@brief A numpy array containing the results of all the candidates
        #
        # The columns in the array correspond to the entries in @a pNAB.pNAB.header
        # and the helical configurations correspond to those in @a pNAB.pNAB.prefix
        self.results = np.loadtxt('results.csv', delimiter=',')

        if self.results.ndim == 1:
            self.results = self.results.reshape((1, len(self.results)))
        ## Add header entries for the dihedral angles; it does not get added in self._run
        for i in range(self.results.shape[1] - 8):
            self.header += ", Dihedral " + str(i+1) + " (degrees)"

        n_candidates = 0 if self.results.size == 0 else len(self.results)

        print("Run is complete. %i out of %i conformers are found"
              %(n_candidates, self._options['RuntimeParameters']['num_candidates']*len(self.prefix)))
        if n_candidates < self._options['RuntimeParameters']['num_candidates']*len(self.prefix):
            print("Maximum number of steps is exceeded.")



# -*- coding: utf-8 -*-

"""Top-level package for pNAB."""

from __future__ import division, absolute_import, print_function

from pnab import bind
from pnab.driver.driver import pNAB
try:
    from pnab.driver.jupyter_widgets import builder
except ImportError:
    pass
from pnab.extras import test



import os

def test(extent='full', extras=None):
    """Runs a test suite through pytest.
    Parameters
    ----------
    extent : {'smoke', 'quick', 'full', 'long'}
        All choices are defined, but choices may be redundant in some projects.
        _smoke_ will be minimal "is-working?" test(s).
        _quick_ will be as much coverage as can be got quickly, approx. 1/3 tests.
        _full_ will be the whole test suite, less some exceedingly long outliers.
        _long_ will be the whole test suite.
    extras : list
        Additional arguments to pass to `pytest`.
    Returns
    -------
    int
        Return code from `pytest.main()`. 0 for pass, 1 for fail.
    """
    try:
        import pytest
    except ImportError:
        raise RuntimeError('Testing module `pytest` is not installed. Run `conda install pytest`.')
    abs_test_dir = os.path.sep.join([os.path.abspath(os.path.dirname(__file__)), "tests"])

    command = ['-rws', '-v']
    if extent.lower() in ['smoke', 'quick', 'full', 'long']:
        pass
    if extras is not None:
        command.extend(extras)
    command.extend(['--capture=sys', abs_test_dir])

    retcode = pytest.main(command)
    return retcode



from __future__ import division, absolute_import, print_function


def test_binder_1():
    """
    test binder

    Test exposed C++ classes attributes and their types
    """
    from pnab import bind

    backbone = bind.Backbone()
    backbone_attr = {'file_path': str, 'interconnects': list, 'linker': list}
    assert all([i in backbone.__dir__() for i in backbone_attr])
    assert all([type(backbone.__getattribute__(k)) is val
                for k, val in backbone_attr.items()])

    base = bind.Base()
    base_attr = {'name': str, 'code': str, 'pair_name': str, 'file_path': str, 'linker': list}
    assert all([i in base.__dir__() for i in base_attr]) 
    assert all([type(base.__getattribute__(k)) is val
                for k, val in base_attr.items()])

    runtime_parameters = bind.RuntimeParameters()
    runtime_parameters_attr = {'energy_filter': list, 'max_distance': float, 'ff_type': str, 'seed': int,
                               'search_algorithm': str, 'num_steps': int, 'dihedral_step': float,
                               'strand': list, 'is_hexad': bool, 'build_strand': list, 'strand_orientation': list,
                               'weighting_temperature': float, 'monte_carlo_temperature': float, 'mutation_rate': float,
                               'crossover_rate': float, 'population_size': int, 'glycosidic_bond_distance': float,
                               'num_candidates': int}
    assert all([i in runtime_parameters.__dir__() for i in runtime_parameters_attr])
    assert all([type(runtime_parameters.__getattribute__(k)) is val
                for k, val in runtime_parameters_attr.items()])

    helical_parameters = bind.HelicalParameters()
    helical_parameters_attr = ['h_twist', 'tip', 'inclination', 'h_rise', 'x_displacement', 'y_displacement',
                               'twist', 'roll', 'tilt', 'rise', 'slide', 'shift',
                               'buckle', 'propeller', 'opening', 'stretch', 'shear', 'stagger']

    assert all([i in helical_parameters.__dir__() for i in helical_parameters_attr])
    assert all([type(helical_parameters.__getattribute__(i)) is float
                for i in helical_parameters_attr])
    assert type(helical_parameters.__getattribute__('is_helical')) is bool


def test_binder_2():
    """
    test binder 2

    Test that the C++ binder runs
    """
    import os

    from pnab import bind

    os.chdir(os.path.dirname(os.path.realpath(__file__)))

    backbone = bind.Backbone()
    backbone.file_path = os.path.join('files', 'rna_bb.pdb')
    backbone.interconnects = [10, 1]
    backbone.linker = [13, 14]

    base = bind.Base()
    base.file_path = os.path.join('files', 'adenine.pdb')
    base.code = 'A'
    base.linker = [5, 11]
    base.name = 'Adenine'
    base.pair_name = 'Uracil'
    bases = [base]

    hp = bind.HelicalParameters()
    hp.h_twist = 32.39
    hp.h_rise = 2.53
    hp.inclination = 22.9
    hp.tip = 0.08
    hp.x_displacement = -4.54
    hp.y_displacement = -0.02

    rp = bind.RuntimeParameters()
    rp.search_algorithm = 'weighted random search'
    rp.num_steps = 1000000
    rp.ff_type = 'GAFF'
    rp.energy_filter = [2.0, 5.0, 10.0, 10000.0, 10000.0]
    rp.max_distance = 0.2
    rp.strand = ['Adenine']*5
    rp.num_candidates = 2

    output = bind.run(rp, backbone, bases, hp, 'test')
    print(output)

def test_helical_parameters():
    """
    Test the equivalence between helical and step parameters.

    The test compares 3 sets of values for helical and step parameters.
    Using the two schemes should give identical results for the accepted candidates.
    To ensure correct comparisons, the step parameter values are set to many digits.
    """
    import os

    from pnab import bind

    os.chdir(os.path.dirname(os.path.realpath(__file__)))

    parameters = [
                  [[1, 2, 3, 4, 5, 6], [-0.4691332223, 0.3125034304, 2.9558478770, -0.5222738076, 0.4178190461, 5.9626387714]],
                  [[-1, -2, -3, -4, -5, -6], [-0.4691332223, 0.3125034304, -2.9558478770, -0.5222738076, 0.4178190461, -5.9626387714]],
                  [[-2, 5, 3, -15, 20, -40], [2.2177307312, 0.4662275467, 4.1753711904, 13.2973947701, 9.9730460776, -36.5122480218]]
                 ]

    backbone = bind.Backbone()
    backbone.file_path = os.path.join('files', 'rna_bb.pdb')
    backbone.interconnects = [10, 1]
    backbone.linker = [13, 14]

    base = bind.Base()
    base.file_path = os.path.join('files', 'adenine.pdb')
    base.code = 'A'
    base.linker = [5, 11]
    base.name = 'Adenine'
    base.pair_name = 'Uracil'
    bases = [base]

    rp = bind.RuntimeParameters()
    rp.search_algorithm = 'weighted random search'
    rp.num_steps = 10
    rp.ff_type = 'GAFF'
    rp.energy_filter = [1e10]*5
    rp.max_distance = 1e10
    rp.strand = ['Adenine']*5
    rp.num_candidates = 10

    for p in parameters:
        hp = bind.HelicalParameters()
        hp.is_helical = True
        hp.x_displacement = p[0][0]
        hp.y_displacement = p[0][1]
        hp.h_rise = p[0][2]
        hp.inclination = p[0][3]
        hp.tip = p[0][4]
        hp.h_twist = p[0][5]

        output_helical = bind.run(rp, backbone, bases, hp, 'test')

        hp.is_helical = False
        hp.shift = p[1][0]
        hp.slide = p[1][1]
        hp.rise = p[1][2]
        hp.tilt = p[1][3]
        hp.roll = p[1][4]
        hp.twist = p[1][5]

        output_step = bind.run(rp, backbone, bases, hp, 'test')

        assert output_helical == output_step



from __future__ import division, absolute_import, print_function

import os
import platform
import glob
import numpy as np

def test_options():
    """
    test parity between the C++ attributes of classes and the python options
    """
    from pnab import bind
    from pnab.driver.options import _options_dict

    components = ['Backbone', 'Base', 'RuntimeParameters', 'HelicalParameters']
    assert all([i in _options_dict for i in components])


    assert all([i in _options_dict['Backbone'] for i in bind.Backbone.__dict__
                if not i.startswith('__')])
    assert all([i in _options_dict['Base'] for i in bind.Base.__dict__
                if not i.startswith('__')])
    assert all([i in _options_dict['RuntimeParameters'] for i in
                bind.RuntimeParameters.__dict__ if not i.startswith('__')])
    assert all([i in _options_dict['HelicalParameters'] for i in
                bind.HelicalParameters.__dict__ if not i.startswith('__')])


def test_examples():
    """
    test running with provided option file
    """
    import pnab

    examples = ['RNA.yaml', 'DNA.yaml', 'FRNA.yaml', 'LNA.yaml', 'CeNA.yaml', 'PNA.yaml', '5methylcytosine.yaml',
                'ZP.yaml', 'Hexad.yaml', 'Hexad_Antiparallel.yaml', 'adenine_cyanuric_acid.yaml']

    if platform.system() != 'Linux': # test times in other platforms exceed allowed limit for travis CI and appveyor
        examples = examples[:7]

    os.chdir(os.path.dirname(os.path.realpath(__file__)))

    for f in examples:
        print("Testing ", f)

        run = pnab.pNAB(f)
        run.run()

        if platform.system() == 'Linux':
            ref_output = np.genfromtxt(os.path.join('files', f.split('.')[0] + '.csv'), delimiter=',')
        elif platform.system() == 'Windows':
            ref_output = np.genfromtxt(os.path.join('files', f.split('.')[0] + '_windows.csv'), delimiter=',')
        elif platform.system() == 'Darwin':
            ref_output = np.genfromtxt(os.path.join('files', f.split('.')[0] + '_mac.csv'), delimiter=',')
            
        assert np.allclose(run.results, ref_output, atol=1e-4)



"""!@file
A file for displaying widgets in the Jupyter notebook

@namespace jupyter_widgets
@brief A file for displaying widgets in the Jupyter notebook

This file containts widgets for sepecifying options in the Jupyter
notebook using the ipywidgets library (https://github.com/jupyter-widgets/ipywidgets).
Using Jupyter notebook is not necessary for using the program, as
the program can be run in as a python script. However, the widgets
provide a graphical user interface for making input files, running the code,
and displaying the results. The Jupyter notebook can be run locally
or on the clouds using Binder (https://mybinder.org/v2/gh/alenaizan/pnab/master?filepath=binder).

The widgets have four main components: Backbone, Bases, Helical Parameters,
and Runtime Parameters. The Bases section does not require the user input
but displays some information about the available nucleobases in the library.
The NGLView library is used for visualization

To run the widgets from a Jupyter notebook, simply execute the following:

@code
import pnab
pnab.builder()
@endcode
"""

import os
import copy
import datetime
from zipfile import ZipFile

import yaml
import ipywidgets as widgets
from IPython.display import display, Javascript, Image

from pnab.driver.driver import pNAB
from pnab.driver.options import _options_dict
from pnab import __path__

##@brief Stores the widgets corresponding to the user-defined options
#
# This dictionary has the same structure as the options._options_dict
# dictionary that has all the available options for the code. 
# The widgets are added to this dictionary and the values of
# the widgets are extracted when the user runs the code
#
# @sa options._options_dict
# @sa extract_options

input_options = {}

def view_nglview(molecule, label=False):
    """!@brief Display molecules using the NGLView viewer.

    A function to view molecules using the NGLView project (https://github.com/arose/nglview).
    It is used to display the geometry of the backbone with atom index labels. It is
    also used to display accepted nucleic acid candidates generated after
    the search. For backbone molecules, the number of atoms in the backbone
    is computed here and returned to be used for bounding the displayed atomic
    indices in backbone widgets.


    @param molecule (str) Path to a file containing the 3D structure of the molecule
    @param label (bool) Whether to display atom index labels

    @return number of atoms in @a molecule if @a label is True, else None

    @attention If the molecules are not displayed correctly, you may need to execute
        @code{.sh} jupyter-nbextension enable nglview --py --sys-prefix @endcode

    @sa jupyter_widgets.path
    @sa jupyter_widgets.single_result
    """

    import openbabel
    import nglview

    # Check if the file exists
    if not os.path.isfile(molecule):
        return 0

    # If a label is requested, then this is a backbone or base molecule
    if label:
        # Read the molecule and convert it to a pdb format
        mol = openbabel.OBMol()
        conv = openbabel.OBConversion()
        fmt = conv.FormatFromExt(molecule)
        conv.SetInAndOutFormats(fmt.GetID(), 'pdb')
        conv.ReadFile(mol, molecule)

        # Extract the number of atoms
        num_atoms = mol.NumAtoms()
        mol = conv.WriteString(mol)

        # Display the molecule
        struct = nglview.TextStructure(mol)
        view = nglview.NGLWidget(struct, defaultRepresentation=False)
        view.camera = 'orthographic'
        view.add_representation('licorice')
        view.add_representation('label', labelType='serial', backgroundColor='black', showBackground=True)
        view.center()
        display(view.display(gui=True))

        # Return the number of atoms to be used for the backbone or base widgets
        return num_atoms

    # Show accepted candidates
    else:
        struct = nglview.FileStructure(molecule)
        view = nglview.NGLWidget(struct, defaultRepresentation=False)
        view.camera = 'orthographic'
        view.add_representation('licorice')
        view.center()
        display(view.display(gui=True))


def fixed_bonds(num_bonds, num_atoms, param):
    """!@brief Display widgets for determining indices of fixed bonds

    This function is changed interactively bases on the number of fixed
    bonds requested by the user.

    @param num_bonds (int) number of fixed bonds; changed dynamically by the user
    @param num_atoms (int) number of atoms in the backbone
    @param param (dict) @a options._options_dict['Backbone']

    @returns None; @a jupyter_widgets.input_options is modified in place

    @sa path 
    @sa input_options
    @sa options._options_dict
    """

    # Add widget to the widget dictionary
    input_options['Backbone']['fixed_bonds'] = []

    # Capture output. This is used to delete widget when no fixed bond is requested
    out = widgets.Output()
    display(out)
    if num_bonds == 0:
        out.clear_output()

    # Loop over the number of requested fixed bonds
    for i in range(num_bonds):
        try:
            # Check to see if values exist in the used input file
            v1, v2 = param['fixed_bonds']['default'][i][0], param['fixed_bonds']['default'][i][1]
        except IndexError:
            # if they don't exist, then set them to 1 and proceed
            v1 = v2 = 1

        # Define two dropdown widgets listing atom indices and a help box
        atom1 = widgets.Dropdown(value=v1, options=range(1, num_atoms + 1), layout=widgets.Layout(width='100px'))
        atom2 = widgets.Dropdown(value=v2, options=range(1, num_atoms + 1), layout=widgets.Layout(width='100px'))
        help_box = widgets.Button(description='?', tooltip=param['fixed_bonds']['long_glossory'], layout=widgets.Layout(width='4%'))
        box = widgets.HBox([help_box, widgets.Label(param['fixed_bonds']['glossory'], layout={'width': '400px'}), atom1, atom2])

        # Display box
        with out:
            display(box)

        # Append new widgets to the widget dictionary
        input_options['Backbone']['fixed_bonds'].append([atom1, atom2])
        

def path(file_path, param):
    """!@brief Display backbone to Jupyter notebook given a file path

    This function displays the options for the backbone. If the @a file_path does
    not exist, this function displays nothing.
    @param file_path (str) Path to a file containing the 3D structure of a backbone molecule
    @param param (dict) @a options._options_dict['Backbone']

    @returns None; @a jupyter_widgets.input_options is modified in place

    @sa upload_backbone
    @sa input_options
    @sa options._options_dict
    @sa view_nglview
    """

    if not os.path.isfile(file_path):
        # Check if the requested file is in the "pnab/data" directory
        file_path = os.path.join(__path__[0], 'data', file_path)
        if not os.path.isfile(file_path):
            # Unset the values for backbone parameters and return
            param['linker']['default'][0] = param['linker']['default'][1] = 1
            param['interconnects']['default'][0] = param['interconnects']['default'][1] = 1
            return

    # Show molecule using NGLView with atom numbers
    num_atoms = view_nglview(file_path, label=True)

    # Display widgets for backbone connection to the nucleobase and the other backbone
    linker1 = widgets.Dropdown(value=param['linker']['default'][0], options=range(1, num_atoms + 1), layout=widgets.Layout(width='100px'))
    linker2 = widgets.Dropdown(value=param['linker']['default'][1], options=range(1, num_atoms + 1), layout=widgets.Layout(width='100px'))
    interconnects1 = widgets.Dropdown(value=param['interconnects']['default'][0], options=range(1, num_atoms + 1), layout=widgets.Layout(width='100px'))
    interconnects2 = widgets.Dropdown(value=param['interconnects']['default'][1], options=range(1, num_atoms + 1), layout=widgets.Layout(width='100px'))

    # Add help boxes
    help_box = widgets.Button(description='?', tooltip=param['linker']['long_glossory'], layout=widgets.Layout(width='4%'))
    box1 = widgets.HBox([help_box, widgets.Label(param['linker']['glossory'], layout={'width': '400px'}), linker1, linker2])

    help_box = widgets.Button(description='?', tooltip=param['interconnects']['long_glossory'], layout=widgets.Layout(width='4%'))
    box2 = widgets.HBox([help_box, widgets.Label(param['interconnects']['glossory'],
                layout={'width': '400px'}), interconnects1, interconnects2])

    # Display
    display(box1)
    display(box2)

    # Add widgets to the widgets dictionary
    input_options['Backbone'] = {'file_path': file_path, 'linker': [linker1, linker2],
                           'interconnects': [interconnects1, interconnects2]}

    help_box = widgets.Button(description='?', tooltip=('This can be used to fix some of the rotatable bonds at their current values' +
                              ' to decrease the number of degrees of freedom in the search.'), layout=widgets.Layout(width='4%'))

    # Display a text widget asking for the number of fixed bonds (default to 0)
    display(widgets.HBox([help_box, widgets.interactive(fixed_bonds, num_bonds=widgets.BoundedIntText(value=len(param['fixed_bonds']['default']), min=0,
                                                        description="Number of fixed rotatable bonds", style={'description_width': 'initial'}),
                                    param=widgets.fixed(param), num_atoms=widgets.fixed(num_atoms))]))



def upload_backbone(f, param):
    """!@brief Upload a backbone file to Jupyter notebook

    This function is used to upload a backbone file and write it. If a file is
    already specified, then the backbone is displayed. If a file is uploaded, this
    function writes the file to the current working directory. Then, the function
    calls the @a jupyter_widgets.path function interactively with the backbone file path. 

    @param f (ipywidgets.FileUpload) File upload widget
    @param param (dict) @a options._options_dict['Backbone']

    @returns None

    @sa path
    @sa backbone
    """
    if f:
        # This is to unset previous values if a new file is provided
        param['linker']['default'][0] = param['linker']['default'][1] = 1
        param['interconnects']['default'][0] = param['interconnects']['default'][1] = 1

        # Get the content of the file in binary format and write it to desk
        input_file = list(f.keys())[0]
        with open(input_file, 'wb') as w:
            w.write(list(f.values())[0]['content'])

        param['file_path']['default'] = input_file

    # Add help box
    help_box = widgets.Button(description='?', tooltip=param['file_path']['long_glossory'], layout=widgets.Layout(width='4%'))

    # Use the backbone file to display the backbone options
    display(widgets.HBox([help_box, widgets.interactive(path, file_path=widgets.Text(value=param['file_path']['default'],
                                        description="Backbone File", style={'description_width': 'initial'}),
                                        param=widgets.fixed(param), upload=widgets.fixed(True))]))


def backbone(param):
    """!@brief Main backbone widget for use in Jupyter notebook

    This function displays the backbone widgets.

    @param param (dict) @a options._options_dict['Backbone']

    @returns None

    @sa upload_backbone
    @sa options._options_dict
    @sa display_options_widgets
    """

    display(widgets.HTML(value='<H3>Backbone</H3>'))

    # Display a widget for uploading backbone files
    w = widgets.interactive(upload_backbone, param=widgets.fixed(param),
                            f=widgets.FileUpload(accept='', multiple=False, 
                            description="Backbone File", style={'description_width': 'initial'}))

    # Add a help box
    help_box = widgets.Button(description='?', tooltip='Upload a backbone file here. The file will be copied to the current folder.',
                              layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, w]))


def base_path(file_path, param, base_number):
    """!@brief Display base to Jupyter notebook given a file path

    This function displays the options for the base. If the @a file_path does
    not exist, this function displays nothing.
    @param file_path (str) Path to a file containing the 3D structure of the base molecule
    @param param (dict) @a options._options_dict['Base']
    @param base_number The number of the new base

    @returns None; @a jupyter_widgets.input_options is modified in place

    @sa upload_base
    @sa input_options
    @sa options._options_dict
    @sa view_nglview
    """

    if not os.path.isfile(file_path):
        # Check if the requested file is in the "pnab/data" directory
        file_path = os.path.join(__path__[0], 'data', file_path)
        if not os.path.isfile(file_path):
            # Unset the values for backbone parameters and return
            param['linker']['default'][0] = param['linker']['default'][1] = 1
            return

    # Show molecule using NGLView with atom numbers
    num_atoms = view_nglview(file_path, label=True)

    # Display widgets for base-to-backbone connection
    linker1 = widgets.Dropdown(value=param['linker']['default'][0], options=range(1, num_atoms + 1), layout=widgets.Layout(width='100px'))
    linker2 = widgets.Dropdown(value=param['linker']['default'][1], options=range(1, num_atoms + 1), layout=widgets.Layout(width='100px'))

    # Add help box
    help_box = widgets.Button(description='?', tooltip=param['linker']['long_glossory'], layout=widgets.Layout(width='4%'))
    box = widgets.HBox([help_box, widgets.Label(param['linker']['glossory'], layout={'width': '400px'}), linker1, linker2])

    # Display
    display(box)

    # Code
    code = widgets.Text(value=param['code']['default'], description=param['code']['glossory'], style={'description_width': 'initial'}, layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['code']['long_glossory'], layout=widgets.Layout(width='4%'))
    box = widgets.HBox([help_box, code])
    display(box)

    # Name
    name = widgets.Text(value=param['name']['default'], description=param['name']['glossory'], style={'description_width': 'initial'}, layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['name']['long_glossory'], layout=widgets.Layout(width='4%'))
    box = widgets.HBox([help_box, name])
    display(box)

    # Pair name
    pair_name = widgets.Text(value=param['pair_name']['default'], description=param['pair_name']['glossory'], style={'description_width': 'initial'}, layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['pair_name']['long_glossory'], layout=widgets.Layout(width='4%'))
    box = widgets.HBox([help_box, pair_name])
    display(box)

    # Align
    align = widgets.Checkbox(value=param['align']['default'], indent=False, description=param['align']['glossory'], style={'description_width': 'initial'}, layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['align']['long_glossory'], layout=widgets.Layout(width='4%'))
    box = widgets.HBox([help_box, align])
    display(box)

    # Add widgets to the widgets dictionary
    input_options['Base %i' %base_number] = {'file_path': file_path, 'linker': [linker1, linker2], 'code': code, 'name': name, 'pair_name': pair_name, 'align': align}


def upload_base(f, param, base_number):
    """!@brief Upload a base file to Jupyter notebook

    This function is used to upload a base file and write it. If a file is
    already specified, then the base is displayed. If a file is uploaded, this
    function writes the file to the current working directory. Then, the function
    calls the @a jupyter_widgets.base_path function interactively with the base file path. 

    @param f (ipywidgets.FileUpload) File upload widget
    @param param (dict) @a options._options_dict['Base'] for the appropriate base number
    @param base_number (int) The number of the additional base

    @returns None

    @sa base_path
    @sa add_base
    """
    if f:
        # This is to unset previous values if a new file is provided
        param['linker']['default'][0] = param['linker']['default'][1] = 1

        # Get the content of the file in binary format and write it to desk
        input_file = list(f.keys())[0]
        with open(input_file, 'wb') as w:
            w.write(list(f.values())[0]['content'])

        param['file_path']['default'] = input_file

    # Add help box
    help_box = widgets.Button(description='?', tooltip=param['file_path']['long_glossory'], layout=widgets.Layout(width='4%'))

    # Use the base file to display the backbone options
    display(widgets.HBox([help_box, widgets.interactive(base_path, file_path=widgets.Text(value=param['file_path']['default'],
                                         description="Base File", style={'description_width': 'initial'}),
                                         param=widgets.fixed(param), upload=widgets.fixed(True), base_number=widgets.fixed(base_number))]))


def add_base(number_of_bases, param):
    """!@brief Display widgets to upload the requested number of bases

    @param number_of_bases (int) The number of additional bases to define
    @param param (dict) @a options._options_dict['Base'] and dictionaries for the other defined bases

    @sa bases
    @sa upload_base
    @sa options._options_dict
    @sa view_nglview
    """
    # Capture output. This is used to delete widget when no additional bases are requested
    out = widgets.Output()
    display(out)
    if number_of_bases == 0:
        out.clear_output()

    for i in range(number_of_bases):
        if "Base %i" %(i+1) not in param:
            param["Base %i" %(i+1)] = copy.deepcopy(param['Base'])
        # Display a widget for uploading base files
        w = widgets.interactive(upload_base, param=widgets.fixed(param["Base %i" %(i+1)]), base_number=widgets.fixed(i + 1),
                                f=widgets.FileUpload(accept='', multiple=False, 
                                description="Base File", style={'description_width': 'initial'}))

        # Add a help box
        help_box = widgets.Button(description='?', tooltip='Upload base file here. The file will be copied to the current folder.',
                                  layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, w]))



def bases(param):
    """!@brief Bases widget for use in Jupyter notebook

    This function displays information on the available nucleobases that are defined in the program.
    The bases are defined in "data/bases_library.yaml"

    @returns None

    @sa display_options_widgets
    """

    display(widgets.HTML(value='<H3>Bases</H3>'))
    display(widgets.HTML(value=('These bases are already defined:' +
                               '<br> Adenine (A), Guanine (G), Cytosine (C), Uracil (U), Thymine (T), ' +
                               'Cyanuric Acid (I), and aminopyrimidine (E)')))

    num_defined_bases = len([k for k in param if 'Base' in k]) - 1 # Subtract the item from the main options
    w = widgets.interactive(add_base, param=widgets.fixed(param), number_of_bases=widgets.BoundedIntText(value=num_defined_bases, min=0,
                                                        description="Number of additional bases", style={'description_width': 'initial'}))

    help_box = widgets.Button(description='?', tooltip=('The number of additional bases to define'), layout=widgets.Layout(width='4%'))
    display(widgets.HBox([help_box, w]))


def helical_parameters(param):
    """
    Display widgets for specifying helical parameters. It also displays an image illustrating
    the used helical parameters.

    @param param (dict) @a options._options_dict['HelicalParameters']

    @returns None; @a jupyter_widgets.input_options is modified in place

    @sa input_options
    @sa display_options_widgets 
    @sa options._options_dict
    """

    display(widgets.HTML(value='<H3>Helical Parameters</H3>'))

    # Display an image showing the helical parameters for DNA and RNA
    display(Image(os.path.join(__path__[0], 'images', 'helical_parameters.jpg'), width=800))

    param_dict = {}
    # Add widget to the widgets dictionary
    input_options['HelicalParameters'] = {}

    def helical_or_step(mode, param):
        if mode == 'Helical':
            excluded_parameters = ['shift', 'slide', 'rise', 'tilt', 'roll', 'twist']
            input_options['HelicalParameters']['is_helical'] =  True
        else:
            excluded_parameters = ['x_displacement', 'y_displacement', 'h_rise', 'inclination', 'tip', 'h_twist']
            input_options['HelicalParameters']['is_helical'] =  False

        for k in param:
            if k == 'is_helical' or k in excluded_parameters:
                continue
            param_dict[k] = []
            default = [param[k]['default'][0], param[k]['default'][1], param[k]['default'][2]] # [beginning point, end point, number of steps]
            # Set angles
            if k in ['inclination', 'tip', 'h_twist', 'roll', 'tilt', 'twist', 'buckle', 'propeller', 'opening']:
                param_dict[k].append(widgets.FloatRangeSlider(value=[default[0], default[1]], min=-180, max=180, step=0.01, readout_format='.2f'))
            # Set distances
            else: # h_rise, x-displacement, y-displacement, rise, slide, shift, stretch, shear, stagger
                # limit maxium and minimum distance values to be between -20 and 20 and 0.01 Angstrom step size
                param_dict[k].append(widgets.FloatRangeSlider(value=[default[0], default[1]], min=-20, max=20, step=0.01, readout_format='.3f'))

            # Add the number of steps widgets
            param_dict[k].append(widgets.BoundedIntText(value=default[2], min=1, max=1000, step=1, description='Steps'))
            help_box = widgets.Button(description='?', tooltip=param[k]['long_glossory'], layout=widgets.Layout(width='3%'))
            box = widgets.HBox([help_box, widgets.Label(param[k]['glossory'], layout={'width': '200px'}), param_dict[k][0], param_dict[k][1]])
            display(box)

            # Add widgets for helical parameter k to the widgets dictionary
            input_options['HelicalParameters'][k] = [param_dict[k][0], param_dict[k][1]]

    default = 'Helical' if param['is_helical']['default'] else 'Step'
    help_box = widgets.Button(description='?', tooltip=param['is_helical']['long_glossory'], layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, widgets.interactive(helical_or_step, mode=widgets.Dropdown(value=default,
                                         options=['Helical', 'Step'],
                                         description="Base step scheme",
                                         style={'description_width': 'initial'},
                                         #layout={'width': '75%'}
                                         ),
                                         param=widgets.fixed(param))]))


def algorithm(chosen_algorithm, param):
    """!@brief Display search parameters based on the chosen algorithm

    There are six search algorithms and each one has a set of input parameters.
    The search algorithms are:

    - Systematic Search: Requires specifying the dihedral angle step size, @a dihedral_step.
    - Monte Carlo Search: Requires specifying the number of steps, @a num_steps, 
        and the Monte Carlo temperature, @a monte_carlo_temperature.
    - Weighted Monte Carlo Search: Also requires specifying the weighting temperature, @a weighting_temperature.
    - Random Search: Requires specifying the number of steps, @a num_steps.
    - Weighted Random Search: Also requires specifying the weighting temperature, @a weighting_temperature.
    - Genetic Algorithm Search: Requires specifying the number of generations, @a num_steps,
        the population size, @a population_size, the mutation_rate, @a mutation_rate,
        and the crossover rate, @a crossover_rate.

    @param chosen_algorithm (str) The chosen algorithm
    @param param (dict) @a options._options_dict['RuntimeParameters']

    @returns None; @a jupyter_widgets.input_options is modified in place

    @sa runtime_parameters
    @sa input_options
    @sa options._options_dict
    """
    
    chosen_algorithm = chosen_algorithm.lower()

    # Add widget to the widgets dictionary
    input_options['RuntimeParameters']['search_algorithm'] = chosen_algorithm

    # Systematic search algorithm needs a dihedral step size
    if chosen_algorithm == 'systematic search':
        dihedral_step = widgets.BoundedFloatText(value=param['dihedral_step']['default'],
                                                 min=0.0,
                                                 max=360.0,
                                                 description=param['dihedral_step']['glossory'],
                                                 style={'description_width': 'initial'},
                                                 layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['dihedral_step']['long_glossory'], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, dihedral_step]))
        input_options['RuntimeParameters']['dihedral_step'] = dihedral_step

    # The other five algorithms require specifying the random number seed and the number of steps or generation
    else:
        # Random number generator seed
        seed = widgets.BoundedIntText(value=param['seed']['default'], min=0, max=2**32-1,
                                      description=param['seed']['glossory'],
                                      style={'description_width': 'initial'},
                                      layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['seed']['long_glossory'], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, seed]))
        input_options['RuntimeParameters']['seed'] = seed

        # Number of steps or generations
        num_steps = widgets.BoundedIntText(value=param['num_steps']['default'], min=1, max=1e100,
                                           description=param['num_steps']['glossory'],
                                           style={'description_width': 'initial'},
                                           layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['num_steps']['long_glossory'], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, num_steps]))
        input_options['RuntimeParameters']['num_steps'] = num_steps

    # If the algorithm weights the dihedral angle probability profile, we need to specify the temperature
    # used in the weighting
    if "weighted" in chosen_algorithm:
        weighting_temperature = widgets.BoundedFloatText(value=param['weighting_temperature']['default'], min=1, max=1e100,
                                                         description=param['weighting_temperature']['glossory'],
                                                         style={'description_width': 'initial'},
                                                         layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['weighting_temperature']['long_glossory'], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, weighting_temperature]))
        input_options['RuntimeParameters']['weighting_temperature'] = weighting_temperature

    # Specify the temperature for the Monte Carlo procedure
    if "monte carlo search" in chosen_algorithm:
        monte_carlo_temperature = widgets.BoundedFloatText(value=param['monte_carlo_temperature']['default'], min=1, max=1e100,
                                                           description=param['monte_carlo_temperature']['glossory'],
                                                           style={'description_width': 'initial'},
                                                           layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['monte_carlo_temperature']['long_glossory'], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, monte_carlo_temperature]))
        input_options['RuntimeParameters']['monte_carlo_temperature'] = monte_carlo_temperature

    # Specify population size, mutation rate, and crossover rate for genetic algorithm
    if 'genetic algorithm search' in chosen_algorithm:
        population_size = widgets.BoundedIntText(value=param['population_size']['default'], min=1, max=1e100,
                                                 description=param['population_size']['glossory'],
                                                 style={'description_width': 'initial'},
                                                 layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['population_size']['long_glossory'], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, population_size]))
        input_options['RuntimeParameters']['population_size'] = population_size

        mutation_rate = widgets.BoundedFloatText(value=param['mutation_rate']['default'], min=0, max=1,
                                                 description=param['mutation_rate']['glossory'],
                                                 style={'description_width': 'initial'},
                                                 layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['mutation_rate']['long_glossory'], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, mutation_rate]))
        input_options['RuntimeParameters']['mutation_rate'] = mutation_rate

        crossover_rate = widgets.BoundedFloatText(value=param['crossover_rate']['default'], min=0, max=1,
                                                  description=param['crossover_rate']['glossory'],
                                                  style={'description_width': 'initial'},
                                                  layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['crossover_rate']['long_glossory'], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, crossover_rate]))
        input_options['RuntimeParameters']['crossover_rate'] = crossover_rate


def runtime_parameters(param):
    """!@brief Runtime parameter widget for use in Jupyter notebook

    Displays widgets for specifying runtime parameters.

    @param param (dict) @a options._options_dict['RuntimeParameters']

    @returns None; @a jupyter_widgets.input_options is modified in place

    @sa algorithm
    @sa input_options
    @sa display_options_widgets 
    @sa options._options_dict
    """

    display(widgets.HTML(value='<H3>Runtime Parameters</H3>'))

    input_options['RuntimeParameters'] = {}

    # Search algorithm
    display(widgets.HTML(value='<H4>Search Algorithm</H4>'))
    dropdown = widgets.Dropdown(value=param['search_algorithm']['default'].title(),
                                options=['Weighted Monte Carlo Search', 'Monte Carlo Search', 'Weighted Random Search', 'Random Search', 'Genetic Algorithm Search', 'Systematic Search'],
                                description=param['search_algorithm']['glossory'],
                                style={'description_width': 'initial'},
                                layout={'width': '75%'})
    search_algorithm = widgets.interactive_output(algorithm, {'chosen_algorithm':dropdown, 'param':widgets.fixed(param)})
    help_box = widgets.Button(description='?', tooltip=param['search_algorithm']['long_glossory'], layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, dropdown]))
    display(search_algorithm)

    # Maximum number of accepted candidates
    num_candidates = widgets.BoundedIntText(value=param['num_candidates']['default'], min=1, max=2**32-1,
                                          description=param['num_candidates']['glossory'],
                                          style={'description_width': 'initial'},
                                          layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['num_candidates']['long_glossory'], layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, num_candidates]))
    input_options['RuntimeParameters']['num_candidates'] = num_candidates

    display(widgets.HTML(value='<H4>Distance and Energy Thresholds</H4>'))

    # Distance and energy thresholds

    # Threshold for the distance
    max_distance = widgets.BoundedFloatText(value=param['max_distance']['default'], min=0.0,
                                            description=param['max_distance']['glossory'],
                                            style={'description_width': 'initial'},
                                            layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['max_distance']['long_glossory'], layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, max_distance]))
    input_options['RuntimeParameters']['max_distance'] = max_distance

    # Force field
    ff_type = widgets.Dropdown(value=param['ff_type']['default'], options=['GAFF', 'MMFF94', 'MMFF94s', 'UFF', 'GHEMICAL'],
                               description=param['ff_type']['glossory'],
                               style={'description_width': 'initial'},
                               layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['ff_type']['long_glossory'], layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, ff_type]))
    input_options['RuntimeParameters']['ff_type'] = ff_type

    # Energy filters
    input_options['RuntimeParameters']['energy_filter'] = []
    for i in range(5):
        minimum = 0 if i < 3 else -1e10 # Put the minimum value to zero for bonded terms
        label = param['energy_filter']['glossory'].split('\n')[i]
        energy_filter = widgets.BoundedFloatText(value=param['energy_filter']['default'][i], min=minimum, max=1e10,
                                                 description=label,
                                                 style={'description_width': 'initial'},
                                                 layout={'width': '75%'})
        help_box = widgets.Button(description='?', tooltip=param['energy_filter']['long_glossory'].split('\n')[i], layout=widgets.Layout(width='3%'))
        display(widgets.HBox([help_box, energy_filter]))
        input_options['RuntimeParameters']['energy_filter'].append(energy_filter)


    display(widgets.HTML(value='<H4>Structural Parameters</H4>'))
    # Base sequence
    strand = widgets.Text(value=''.join(param['strand']['default']),
                          description=param['strand']['glossory'],
                          style={'description_width': 'initial'},
                          layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['strand']['long_glossory'], layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, strand]))
    input_options['RuntimeParameters']['strand'] = strand

    # Pair adenine with uracil? Default is A-T base pair
    pair_A_U = widgets.Checkbox(value=param['pair_A_U']['default'], indent=False,
                                description=param['pair_A_U']['glossory'],
                                style={'description_width': 'initial'},
                                layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['pair_A_U']['long_glossory'], layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, pair_A_U]))
    input_options['RuntimeParameters']['pair_A_U'] =  pair_A_U

    # Is hexad
    is_hexad = widgets.Checkbox(value=param['is_hexad']['default'], indent=False,
                                description=param['is_hexad']['glossory'],
                                style={'description_width': 'initial'},
                                layout={'width': '75%'})
    help_box = widgets.Button(description='?', tooltip=param['is_hexad']['long_glossory'], layout=widgets.Layout(width='3%')) 
    display(widgets.HBox([help_box, is_hexad]))
    input_options['RuntimeParameters']['is_hexad'] =  is_hexad

    # Build strand
    input_options['RuntimeParameters']['build_strand'] = []
    help_box = widgets.Button(description='?', tooltip=param['build_strand']['long_glossory'], layout=widgets.Layout(width='3%'))
    box = [help_box, widgets.Label(param['build_strand']['glossory'])]
    for i in range(6):
        build_strand = widgets.Checkbox(value=param['build_strand']['default'][i], indent=False, layout={'width': '50px'})
        input_options['RuntimeParameters']['build_strand'].append(build_strand)
        box.append(build_strand)
    box = widgets.HBox(box, layout={'width':'100%'})
    display(box)

    # Orientation of each strand in the hexad
    input_options['RuntimeParameters']['strand_orientation'] = []
    help_box = widgets.Button(description='?', tooltip=param['strand_orientation']['long_glossory'], layout=widgets.Layout(width='3%'))
    box = [help_box, widgets.Label(param['strand_orientation']['glossory'])]
    for i in range(6):
        strand_orientation = widgets.Checkbox(value=param['strand_orientation']['default'][i], indent=False, layout={'width': '50px'})
        input_options['RuntimeParameters']['strand_orientation'].append(strand_orientation)
        box.append(strand_orientation)
    box = widgets.HBox(box, layout={'width':'100%'})
    display(box)


def upload_input(param, f):
    """!@brief Widget to upload an input file

    If a file is uploaded, it is written in the current working directory.
    Then, this function calls @a jupyter_widgets.display_options_widgets function with the newly
    written file.

    @param param (dict) @a options._options_dict
    @param f (ipywidgets.FileUpload) file upload widget

    @returns None

    @sa display_options_widgets
    """
    if f:
        # If an input file is provided, write it to the current working directory
        input_file = list(f.keys())[0]
        with open(input_file, 'wb') as w:
            w.write(list(f.values())[0]['content'])

        # Display the name of the input file
        display(widgets.HTML(value=input_file))

        # Display all the widgets with the options specified in the input file
        display_options_widgets(param, input_file, uploaded=True)


def display_options_widgets(param, input_file, uploaded=False):
    """!@brief Display all widgets in Jupyter notebook given an input file

    If @a input_file is provided, then it updates the default options in
    @a options._options_dict to display the user-defined options. If @a
    input_file is "Upload file", then an ipywidgets.FileUpload widget is displayed.
    This function also displays a widget for running the program after the
    user defines all the options.


    @param param (dict) @a options._options_dict
    @param input_file (str) Input file
    @param uploaded (bool) Whether a file is uploaded by the user or not

    @returns None

    @sa backbone
    @sa bases
    @sa helical_parameters
    @sa runtime_parameters
    @sa upload_input
    @sa run
    @sa user_input_file
    @sa options._options_dict
    """

    if input_file == "Upload file":
        # Display a widget for uploading input file
        w = widgets.interactive(upload_input, param=widgets.fixed(param), f=widgets.FileUpload(accept='', multiple=False, description="Input File"))
        help_box = widgets.Button(description='?', tooltip='Upload your input file here. The file will be copied to the current folder.', layout=widgets.Layout(width='4%'))
        display(widgets.HBox([help_box, w]))
        return

    # If the file is not uploaded, then we are using one of the example files
    # in the pnab/data directory
    if not uploaded:
        input_file = os.path.join(__path__[0], 'data', input_file)

    options = yaml.load(open(input_file, 'r'), yaml.FullLoader)

    # Clean dictionary from additional bases if they are not defined in the options
    # This is necessary when the user switches from an input file that has additional bases to one that does not
    num_defined_bases = len([k for k in param if 'Base' in k]) - 1 # Subtract the item from the main options
    for i in range(1, num_defined_bases + 1):
        param.pop('Base %i' %i)
        input_options.pop('Base %i' %i)

    # Update the default options to display those provided by the user
    for k1, v1 in options.items():
        if 'Base' in k1:
            param[k1] = copy.deepcopy(param['Base'])
        for k2, v2 in options[k1].items():
            param[k1][k2]['default'] = options[k1][k2]

    # Display all options widgets
    backbone(param['Backbone'])
    bases_param = {k:val for k, val in param.items() if 'Base' in k}
    bases(bases_param)
    helical_parameters(param['HelicalParameters'])
    runtime_parameters(param['RuntimeParameters'])

    # Display run widget
    button = widgets.Button(description='Run', tooltip='Click here to run the program with the provided options. Once the program finishes, the results will be displayed below.')
    button.on_click(run)
    display(button)


def user_input_file(param):
    """!@brief Display input file options

    Gives the user several input files that they can try, and the option that
    they upload their input file. Once an input file is chosen, all widgets
    are displayed with the user-defined options.

    @param param (dict) @a options._options_dict

    @returns None

    @sa display_options_widgets
    """

    # Provide input files as examples
    examples = ['RNA.yaml', 'DNA.yaml', 'FRNA.yaml', 'LNA.yaml', 'CeNA.yaml', 'PNA.yaml', '5methylcytosine.yaml',
                'ZP.yaml', 'Hexad.yaml', 'adenine_cyanuric_acid.yaml', 'Upload file']
    w = widgets.interactive(display_options_widgets, param=widgets.fixed(param), uploaded=widgets.fixed(False),
            input_file=widgets.Dropdown(options=examples,
                                        style={'description_width': 'initial'}, description='Input File'))
    help_box = widgets.Button(description='?', tooltip=('There are existing example files for RNA, DNA, and other nucleic acid analogs.' +
                                                        ' You can use these examples as a starting point for customizing your input options.' + 
                                                        ' Alternatively, you can upload your own input file.'),
               layout=widgets.Layout(width='3%'))
    display(widgets.HBox([help_box, w]))


def run(button):
    """!@brief Function to run the code when the user finishes specifying all options

    It extracts the user-defined options and creates a @a pNAB.pNAB instance. Then,
    it run the code and display the results.

    @param button (variable) variable for using the widgets.Button.on_click method

    @sa display_options_widgets
    @sa pNAB.pNAB
    @sa show_results
    """

    # Extract options
    run_options = extract_options()

    # Write a file named "options.yaml" containing all the user-defined options
    file_path = 'options.yaml'

    # Make sure we do not overwrite other files that have the same name
    # by prepending enough "_"
    while True:
        if os.path.isfile(file_path):
            file_path = '_' + file_path
        else:
            if os.path.isfile('options.yaml'):
                os.rename('options.yaml', file_path)
            break

    # Write time stamps
    time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    with open('options.yaml', 'w') as f:
        f.write('# ' + time + '\n')
        f.write(yaml.dump(run_options))

    # Run the code
    run = pNAB('options.yaml')
    run.run(interrupt=True)

    # If no results are found, return
    if run.results.size == 0:
        return

    # Get output files
    files = [str(int(conformer[0])) + '_' + str(int(conformer[1])) + '.pdb' for conformer in run.results]

    time = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
    with ZipFile('output' + time + '.zip', 'w') as z:
        for f in ['options.yaml', 'results.csv', 'prefix.yaml']:
            z.write(f)
        for f in files:
            z.write(f)

    display(widgets.HTML("""<a href="output""" + time + """.zip" target="_blank">Download Output</a>"""))
    display(Javascript("""var url="output%s.zip"\nwindow.open(url, 'download')""" %time))

    # display conformers and their properties
    # Sort by total energy
    run.results = run.results[run.results[:, 7].argsort()]
    show_results(run.results, run.header, run.prefix)


def extract_options():
    """!@brief Extracts user options from the widgets

    This function extracts the values of all options specifed by the 
    user. It access all the widgets in @a jupyter_widgets.input_options
    and extracts their values.

    @returns user_options A dictionary of the values of all the user-defined options

    @sa run
    @sa input_options
    @sa options._options_dict
    """

    user_options = {}
    # I did not use a smart way to get the values, but this should work
    for k1, val1 in input_options.items():
        user_options[k1] = {}
        for k2, val2 in val1.items():
            if isinstance(val2, str):
                user_options[k1][k2] = val2
            elif k1 == 'Backbone' or 'Base' in k1:
                if k2 == 'fixed_bonds':
                    user_options[k1][k2] = []
                    for l in val2:
                        user_options[k1][k2].append([l[0].value, l[1].value])
                elif isinstance(val2, list):
                    user_options[k1][k2] = [val2[0].value, val2[1].value]
                elif isinstance(val2, str):
                    user_options[k1][k2] = val2
                else:
                    user_options[k1][k2] = val2.value

            elif k1 == 'HelicalParameters':
                if k2 == 'is_helical':
                    user_options[k1][k2] = val2
                else:
                    user_options[k1][k2] = [val2[0].value[0], val2[0].value[1], val2[1].value]

            else:
                if k2 == 'energy_filter' or k2 == 'strand_orientation' or k2 == 'build_strand':
                    user_options[k1][k2] = [i.value for i in val2]

                else:
                    user_options[k1][k2] = val2.value

    return user_options


def single_result(result, header, results, prefix):
    """!@brief Interactive function to display a single result

    It displays one accepted conformer and prints its energies and other properties.

    @param result (int) index of the conformer in the @a results array
    @param header (str) A comma separated string of the output properties of the conformer
    @param results (np.ndarray) a numpy array of all the accepted conformers
    @param prefix (dict) A dictionary of the prefix of the run and the associated helical configuration

    @returns None    

    @sa show_results
    """

    # Extract result from the array
    result = results[result]

    # Get the name of the PDB file containing the conformer
    conformer = str(int(result[0])) + '_' + str(int(result[1])) + '.pdb'

    # Print information
    print(conformer)
    print("Helical Parameter Set %i: " %result[0], prefix['%i' %result[0]])

    for i in range(2, len(result)):
        print(header.split(', ')[i] + ': %.3f' %result[i])

    # Display the molecule
    view_nglview(conformer)
    

def show_results(results, header, prefix):
    """!@brief Display results

    A function to display all the accepted candidates using a dropdown list.

    @param results (np.ndarray) numpy array of results sorted by total energy
    @param header (str) A comma separated string of the output properties of the conformer
    @param prefix (dict) A dictionary of the prefix of the run and the associated helical configuration

    @returns None

    @sa run
    @sa single_result 
    """

    # Set a list of conformer names and their indices in the results
    options = [(str(int(conformer[0])) + '_' + str(int(conformer[1])) + '.pdb', i) for i, conformer in enumerate(results)]

    # Show a dropdown widget of all the accepted conformers
    dropdown = widgets.Dropdown(value=options[0][1], options=options, style={'description_width': 'initial'}, description='Conformer')
    w = widgets.interactive(single_result, result=dropdown, header=widgets.fixed(header), results=widgets.fixed(results), prefix=widgets.fixed(prefix))
    display(w)



def builder():
    """!@brief The function called from the Jupyter notebook to display the widgets

    Execute the following to display all the widgets in the notebook
    @code
    import pnab
    pnab.builder()
    @endcode

    @returns None
    """

    # Prevents auto-scrolling of the notebook
    disable_js = """
IPython.OutputArea.prototype._should_scroll = function(lines) {
    return false;
}
"""
    display(Javascript(disable_js))

    user_input_file(_options_dict)



"""!@file
A file for preparing, explaining and validating options

@namespace options
@brief A file for preparing, explaining and validating options

This option file defines a dictionary containing all the available options for the builder code.
@a options._options_dict has three keys: Backbone, HelicalParameters, and RuntimeParameters.
Each of these contains a dictionary of available options in each category. Each dictionary contains
a short glossory that describe the option, a long glossory giving hints on how to
use this option, a default value, and a validation scheme.

This file also contains additional functions to help in the validation.
"""

from __future__ import division, absolute_import, print_function

import os
from openbabel import openbabel as ob
import yaml

from pnab import __path__


def validate_all_options(options):
    """!@brief A method to validate all options.

    It loops over all available builder options and validates the user input. For options
    that have not been specified by the user, a default value is provided.

    @param options (dict) A dictionary containing all the user-defined options

    @return None; @a options is modified in place

    @sa pNAB.pNAB.__init__
    @sa _options_dict
    """

    # Loop over all available options
    for k1 in _options_dict:
        if k1 not in options:
            if k1 == 'Base': continue
            options[k1] = {}
        for k2 in _options_dict[k1]:
            if k2 not in options[k1]:
                # Add a default value for options not specified by the user
                options[k1][k2] = _options_dict[k1][k2]['default']
            options[k1][k2] = _options_dict[k1][k2]['validation'](options[k1][k2])

    # Validate base options if additional bases are specified
    for k1 in options:
        if 'Base' in k1:
            for k2 in _options_dict['Base']:
                if k2 not in options[k1]:
                     # Add a default value for options not specified by the user
                      options[k1][k2] = _options_dict['Base'][k2]['default']
                options[k1][k2] = _options_dict['Base'][k2]['validation'](options[k1][k2])
            if options[k1].pop('align'):
                _align_nucleobase(options[k1])

    # Validate whether some options are mutually exclusive
    # Cannot build hexad strands for canonical nucleobases
    if options['RuntimeParameters']['is_hexad']:
        for i in ['A', 'G', 'C', 'U', 'T']:
            if i in options['RuntimeParameters']['strand']:
                raise Exception("Cannot build hexads for canonical nucleobases")

def _align_nucleobase(base_options):
    """!@brief Aligns the provided nucleobase to purine or pyrimidine in the nucleic acid base pair standard reference frame

    This function aligns the provided nucleobase to either guanine if it has two rings or to uracil if it has one ring.
    The provided linker atoms that connect the base to the backbone are used for alignment. A third atom used in the alignment is 
    determined by rearraning the atoms in the molecules using the canonical order. The function loops over all atoms in the provided
    nucleobase and the reference nucleobase and if two atoms have the same atomic number then these atoms are used for the alignment.
    A new file with the aligned nucleobase is created and used instead of the provided file.

    @param base_options (dict) The nucleobase options defined in the input file
    """

    conv = ob.OBConversion()
    nucleobase = ob.OBMol()
    # Read the provided nucleobase
    conv.ReadFile(nucleobase, base_options['file_path'])
    rings = list(nucleobase.GetSSSR())
    # Use either guanine or uracil as a reference
    if len(rings) == 2:
        reference = 'G'
    elif len(rings) == 1:
        reference = 'U'
    else:
        return

    # Read the data for the reference nucleobase
    ref_library = yaml.load(open(os.path.join(__path__[0], 'data', 'bases_library.yaml')), yaml.FullLoader) 
    ref_linker = ref_library["Base " + reference]['linker']
    ref_path = ref_library["Base " + reference]['file_path']

    ref = ob.OBMol()
    conv.ReadFile(ref, os.path.join(__path__[0], 'data', ref_path))

    # Find the ring atoms in the reference base
    # compute the centroid
    centroid_ref = ob.vector3()
    num_atoms = 0
    ref_ring = ob.OBMol()
    # Add the two linker atoms
    ref_ring.AddAtom(ref.GetAtom(ref_linker[1]))
    ref_ring.AddAtom(ref.GetAtom(ref_linker[0]))
    ref_ring.AddBond(1, 2, 1)
    for atom in ob.OBMolAtomIter(ref):
        if atom.IsInRing() and atom.GetIdx() != ref_linker[0]:
            ref_ring.AddAtom(atom)
            centroid_ref += atom.GetVector()
            num_atoms += 1
    centroid_ref /= num_atoms

    # Find the ring atoms in the provided nucleobase
    # compute the centroid
    centroid_nucleobase = ob.vector3()
    num_atoms = 0 
    nucleobase_ring = ob.OBMol()
    # Add the two linker atoms
    nucleobase_ring.AddAtom(nucleobase.GetAtom(base_options['linker'][1]))
    nucleobase_ring.AddAtom(nucleobase.GetAtom(base_options['linker'][0]))
    nucleobase_ring.AddBond(1, 2, 1)
    for atom in ob.OBMolAtomIter(nucleobase):
        if atom.IsInRing() and atom.GetIdx() != base_options['linker'][0]:
            nucleobase_ring.AddAtom(atom)
            centroid_nucleobase += atom.GetVector()
            num_atoms += 1
    centroid_nucleobase /= -1*num_atoms

    # Pick three atoms for alignment
    # First, add the two linker atoms
    ref_align = ob.OBMol()
    nucleobase_align = ob.OBMol()
    for i in range(1, 3):
        ref_align.AddAtom(ref_ring.GetAtom(i))
        nucleobase_align.AddAtom(nucleobase_ring.GetAtom(i))

    # Reorder the atoms using the canonical order
    canonical = ob.OBOp.FindType("canonical")
    canonical.Do(ref_ring)
    canonical.Do(nucleobase_ring)

    # Delete the two linker atoms
    for atom in ob.OBAtomAtomIter(ref_ring.GetAtom(1)):
        ref_ring.DeleteAtom(atom)
    ref_ring.DeleteAtom(ref_ring.GetAtom(1))

    for atom in ob.OBAtomAtomIter(nucleobase_ring.GetAtom(1)):
        nucleobase_ring.DeleteAtom(atom)
    nucleobase_ring.DeleteAtom(nucleobase_ring.GetAtom(1))

    # Search for two atoms that have the same index and atomic number
    for i in range(1, ref_ring.NumAtoms() + 1):
        if ref_ring.GetAtom(i).GetAtomicNum() == nucleobase_ring.GetAtom(i).GetAtomicNum():
            ref_align.AddAtom(ref_ring.GetAtom(i))
            nucleobase_align.AddAtom(nucleobase_ring.GetAtom(i))
            break

    # Align the three atoms in the reference nucleobase and the provided nucleobase
    align = ob.OBAlign(ref_align, nucleobase_align, True, True)
    align.Align()

    # Apply the rotation to all the atoms
    matrix = align.GetRotMatrix()
    array = ob.doubleArray(9)
    matrix.GetArray(array)
    nucleobase.Translate(centroid_nucleobase)
    nucleobase.Rotate(array)
    nucleobase.Translate(centroid_ref)

    # Write a new file with the aligned coordinates and update the options
    base_options['file_path'] = os.path.splitext(base_options['file_path'])[0] + '_aligned' + os.path.splitext(base_options['file_path'])[1]
    conv.WriteFile(nucleobase, base_options['file_path'])

def _validate_input_file(file_name):
    """!@brief Method to validate that the given file exists.

    Check whether the file exists or not. If it does not exist as it is, the function checks
    whether the file is in the "pnab/data" directory. 

    @param file_name (str) Path to a file

    @return file_name after validation

    @sa validate_all_options
    @sa _options_dict
    """

    file_name = str(file_name)
    if not os.path.isfile(file_name):
        # Check if the file exists in the "pnab/data" directory
        file_name2 = os.path.join(__path__[0], 'data', file_name)
        if not os.path.isfile(file_name2):
            raise Exception("Cannot find file: %s" %file_name)
        file_name = file_name2

    return file_name


def _validate_atom_indices(indices):
    """!@brief Method to validate provided lists of atom indices.

    Check whether the list has two atom indices

    @param indices (str|list) atom indices

    @return indices after validation

    @sa validate_all_options
    @sa _options_dict
    """

    indices = list(eval(indices)) if isinstance(indices, str) else list(indices)
    if len(indices) != 2:
        raise Exception("Incorrect number of atoms. Must provide two indices.")

    for i in range(len(indices)):
        indices[i] = int(indices[i])
        if indices[i] == 0:
            raise Exception("Use 1-based index")

    return indices


def _validate_helical_parameters(hp_i):
    """!@brief Method to validate provided helical parameters for each parameter.

    Check whether the list has correct helical parameter specifications.
    The correct specifications are [initial value in a range, final value in a range, number of steps].
    If a single value is provided, then we assume it is the only value in a range.
    if two values are provided, then we assume it is a range with one configuration.

    This specification is for internal use in the driver for generating multiple configurations
    and running them in parallel. The C++ code accepts a single value for each helical parameter.

    @param  hp_i (str|float|list) Specifications for helical parameter i

    @return hp_i after validation

    @sa validate_all_options
    @sa _options_dict
    @sa pNAB.pNAB.run
    """

    if isinstance(hp_i, str):
        hp_i = eval(hp_i)

    # If it is a single int or float value
    if isinstance(hp_i, (float, int)):
        hp_i = [hp_i, hp_i, 1]
    # if it is a list
    else:
        hp_i = list(hp_i)

    # If a list has one value, then we assume it is a single value for the helical parameters
    if len(hp_i) == 1:
        hp_i = [hp_i[0], hp_i[0], 1]
    # If the list has two values, then we assume it is a range with one generated helical configuration
    elif len(hp_i) == 2:
        # Default of 1 value tested for a range
        hp_i.append(1)
    elif len(hp_i) > 3:
        raise Exception("Helical parameters must have at most three values")

    return hp_i


def _validate_energy_filter(energy_filter):
    """!@brief Method to validate provided energy filter.

    Energy filter has five thresholds [bond, angle, torsion, van der Waals, total]

    @param energy_filter (str|list) A list of five energy thresholds

    @return energy_filter after validation

    @sa validate_all_options
    @sa _options_dict
    """

    if isinstance(energy_filter, str):
        energy_filter = eval(energy_filter)

    energy_filter = list(energy_filter)
    if len(energy_filter) != 5:
        raise Exception("Five values must be provided for the energy filter")

    for i in energy_filter:
        if not isinstance(i, (float, int)):
            raise Exception("Provide valid numbers")

    return energy_filter

def _validate_strand(strand):
    """!@brief Method to validate provided strand sequence.

    We use FASTA strings to specify the sequence in the strand.
    This is for use in the driver. The C++ code accepts base
    names with more than one letter. However, writing a FASTA
    sequence is easier.

    The names of the bases are converted to upper case letters

    @param strand (str|list) FASTA sequence of the strand

    @return strand list after validation

    @sa validate_all_options
    @sa _options_dict
    """

    strand = list(strand)
    for i in range(len(strand)):
        strand[i] = strand[i].upper()
    # Check if both canonical and noncanonical bases are in the sequence
    if "I" in strand or "E" in strand:
        for i in ["A", "G", "C", "T", "U"]:
            if i in strand:
                raise Exception("Cannot combine canonincal and non-canonical nucleobases")

    return strand

def _validate_bool_list(x):
    """!@brief Method to validate a list of bools

    @param x (list) list of bools

    @return validated list of bools

    @sa validate_all_options
    @sa _options_dict
    """

    x = list(x)
    for i, val in enumerate(x):
        x[i] = bool(eval(val.title())) if isinstance(val, str) else bool(val)

    return x


# Set glossory of options, default values and validation methods

## @brief Nucleic acid builder options used in the driver.
#
# This is an internal options dictionary to be used for listing, explaining
# and validating the options used in the code. All options here have corresponding
# options in the C++ code, except @a options._options_dict['RuntimeParameters']['pair_a_u'],
# which is used in the driver to override the default adenine-thymine pairing.
# @sa jupyter_widgets.builder()
# @sa PNAB::Backbone
# @sa PNAB::HelicalParameters
# @sa PNAB::RuntimeParameters
_options_dict = {}

# Backbone Parameter
_options_dict['Backbone'] = {}
_options_dict['Backbone']['file_path'] = {
                                         'glossory': ('Path to the file containing the molecular' +
                                                     ' structure of the backbone'),
                                         'long_glossory': ('Path to the file containing the three-dimensional' + 
                                                          ' molecular structure of the backbone (e.g. PDB file).' + 
                                                          ' The backbone must contain hydrogen atoms and not contain' +
                                                          ' atoms from the nucleobase. The atoms included here must' + 
                                                          ' be for one nucleotide.'),
                                         'default': '',
                                         'validation': lambda x: _validate_input_file(x),
                                         }
_options_dict['Backbone']['interconnects'] = {
                                             'glossory': 'Two atoms connecting to the two backbones',
                                             'long_glossory': ('Select two atoms that connect the backbone molecule to' +
                                                               ' the two neighboring backbone molecules. Extra hydrogen atoms' + 
                                                               ' connected to these two linker atoms are deleted. The order of' + 
                                                               ' atoms control the directionality of the backbone.'),
                                             'default': None,
                                             'validation': lambda x: _validate_atom_indices(x),
                                             }
_options_dict['Backbone']['linker'] = {
                                      'glossory': ('Two atoms forming the vector' + 
                                                  ' connecting to base'),
                                      'long_glossory': ('Select two atoms that form the vector connecting the backbone molecule to' +
                                                        ' the nucleobase. The terminal atom will be deleted as the bond with the nucleobase' +
                                                        ' is formed.'),
                                      'default': None,
                                      'validation': lambda x: _validate_atom_indices(x),
                                      }
_options_dict['Backbone']['fixed_bonds'] = {
                                           'glossory': ('Indices of fixed rotatable dihedral bonds'),
                                           'long_glossory': ('Select the indices of the two atoms that are at the center' + 
                                                             ' of the rotatable dihedral angle.'),
                                           'default': [],
                                           'validation': lambda x: [_validate_atom_indices(i) for i in list(x)],
                                           }

# Base Parameters
_options_dict['Base'] = {}

_options_dict['Base']['file_path'] = {
                                       'glossory': 'Path to file containing the molecular structure of the base',
                                       'long_glossory': ('Path to the file containing the three-dimensional' +
                                                        ' molecular structure of the nucleobase (e.g. PDB file).' +
                                                        ' The base must contain hydrogen atoms and not contain' +
                                                        ' atoms from the backbone.'),
                                       'default': '',
                                       'validation': lambda x: _validate_input_file(x),
                                       }
_options_dict['Base']['linker'] = {
                                    'glossory': 'Two atoms forming a vector connecting to backbone',
                                    'long_glossory': ('Select two atoms that form the vector connecting the nucleobase molecule to' +
                                                      ' the backbone. The terminal atom will be deleted as the bond with the backbone' +
                                                      ' is formed.'),
                                    'default': [0, 0],
                                    'validation': lambda x: _validate_atom_indices(x),
                                    }
_options_dict['Base']['code'] = {
                                  'glossory': 'Three-letter code',
                                  'long_glossory': 'This code is used as the residue name in the output PDB files.',
                                  'default': 'RES',
                                  'validation': lambda x: str(x),
                                  }
_options_dict['Base']['name'] = {
                                  'glossory': 'One-letter base name',
                                  'long_glossory': ('This name is used when specifying the strand sequence. It must not be one of the' +
                                                    ' names defined in the program library (A, G, C, T, U, I, E).'),
                                  'default': 'R',
                                  'validation': lambda x: str(x),
                                  }
_options_dict['Base']['pair_name'] = {
                                       'glossory': 'Name of the pairing base',
                                       'long_glossory': 'The one-letter name of the pairing base.',
                                       'default': '',
                                       'validation': lambda x: str(x),
                                       }
_options_dict['Base']['align'] = {
                                 'glossory': 'Align the base to the standard frame of reference',
                                 'long_glossory': ('Aligns the provided nucleobase to either purine or pyrimidine in the base pair standard frame of reference.' +
                                                   ' This alignment is not available for the hexad reference frame.'),
                                 'default': False,
                                 'validation': lambda x: bool(x),
                                 }

# Helical Parameters
_options_dict['HelicalParameters'] = {}
_options_dict['HelicalParameters']['x_displacement'] = {
                                                       'glossory': 'X-Displacement (Angstrom)',
                                                       'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                                         ' you can control the number of configurations generated in that range by increasing' +
                                                                         ' the number of steps. Uniformly spaced in the provided range will be generated.' +
                                                                         ' Not defined for the hexad geometry.'),
                                                       'default': [0.0, 0.0, 1],
                                                       'validation': lambda x: _validate_helical_parameters(x),
                                                       }
_options_dict['HelicalParameters']['y_displacement'] = {
                                                       'glossory': 'Y-Displacement (Angstrom)',
                                                       'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                                         ' you can control the number of configurations generated in that range by increasing' +
                                                                         ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                                         ' Not defined for the hexad geometry.'),
                                                       'default': [0.0, 0.0, 1],
                                                       'validation': lambda x: _validate_helical_parameters(x),
                                                       }
_options_dict['HelicalParameters']['h_rise'] = {
                                                'glossory': 'Helical Rise (Angstrom)',
                                                'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                                  ' you can control the number of configurations generated in that range by increasing' +
                                                                  ' the number of steps. Uniformly spaced configurations in the provided range will be generated.'),
                                                'default': [0.0, 0.0, 1],
                                                'validation': lambda x: _validate_helical_parameters(x),
                                                }
_options_dict['HelicalParameters']['inclination'] = {
                                                    'glossory': 'Inclination (degree)',
                                                    'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                                      ' you can control the number of configurations generated in that range by increasing' +
                                                                      ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                                      ' Not defined for the hexad geometry.'),
                                                    'default': [0.0, 0.0, 1],
                                                    'validation': lambda x: _validate_helical_parameters(x),
                                                    }
_options_dict['HelicalParameters']['tip'] = {
                                            'glossory': 'Tip (degree)',
                                            'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                              ' you can control the number of configurations generated in that range by increasing' +
                                                              ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                              ' Not defined for the hexad geometry.'),
                                            'default': [0.0, 0.0, 1],
                                            'validation': lambda x: _validate_helical_parameters(x),
                                            }
_options_dict['HelicalParameters']['h_twist'] = {
                                               'glossory': 'Helical Twist (degree)',
                                               'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                                 ' you can control the number of configurations generated in that range by increasing' +
                                                                 ' the number of steps. Uniformly spaced configurations in the provided range will be generated.'),
                                               'default': [0.0, 0.0, 1],
                                               'validation': lambda x: _validate_helical_parameters(x),
                                               }


_options_dict['HelicalParameters']['shift'] = {
                                              'glossory': 'Shift (Angstrom)',
                                              'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                                ' you can control the number of configurations generated in that range by increasing' +
                                                                ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                                ' Not defined for the hexad geometry.'),
                                              'default': [0.0, 0.0, 1],
                                              'validation': lambda x: _validate_helical_parameters(x),
                                              }
_options_dict['HelicalParameters']['slide'] = {
                                              'glossory': 'Slide (Angstrom)',
                                              'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                                ' you can control the number of configurations generated in that range by increasing' +
                                                                ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                                ' Not defined for the hexad geometry.'),
                                              'default': [0.0, 0.0, 1],
                                              'validation': lambda x: _validate_helical_parameters(x),
                                              }
_options_dict['HelicalParameters']['rise'] = {
                                             'glossory': 'Rise (Angstrom)',
                                             'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                               ' you can control the number of configurations generated in that range by increasing' +
                                                               ' the number of steps. Uniformly spaced configurations in the provided range will be generated.'),
                                             'default': [0.0, 0.0, 1],
                                             'validation': lambda x: _validate_helical_parameters(x),
                                             }
_options_dict['HelicalParameters']['tilt'] = {
                                             'glossory': 'Tilt (degree)',
                                             'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                               ' you can control the number of configurations generated in that range by increasing' +
                                                               ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                               ' Not defined for the hexad geometry.'),
                                             'default': [0.0, 0.0, 1],
                                             'validation': lambda x: _validate_helical_parameters(x),
                                             }
_options_dict['HelicalParameters']['roll'] = {
                                             'glossory': 'Roll (degree)',
                                             'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                               ' you can control the number of configurations generated in that range by increasing' +
                                                               ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                               ' Not defined for the hexad geometry.'),
                                             'default': [0.0, 0.0, 1],
                                             'validation': lambda x: _validate_helical_parameters(x),
                                             }
_options_dict['HelicalParameters']['twist'] = {
                                              'glossory': 'Twist (degree)',
                                              'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                                ' you can control the number of configurations generated in that range by increasing' +
                                                                ' the number of steps. Uniformly spaced configurations in the provided range will be generated.'),
                                              'default': [0.0, 0.0, 1],
                                              'validation': lambda x: _validate_helical_parameters(x),
                                              }


_options_dict['HelicalParameters']['shear'] = {
                                            'glossory': 'Shear (Angstrom)',
                                            'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                              ' you can control the number of configurations generated in that range by increasing' +
                                                              ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                              ' Not defined for the hexad geometry.'),
                                            'default': [0.0, 0.0, 1],
                                            'validation': lambda x: _validate_helical_parameters(x),
                                            }
_options_dict['HelicalParameters']['stretch'] = {
                                            'glossory': 'Stretch (Angstrom)',
                                            'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                              ' you can control the number of configurations generated in that range by increasing' +
                                                              ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                              ' Not defined for the hexad geometry.'),
                                            'default': [0.0, 0.0, 1],
                                            'validation': lambda x: _validate_helical_parameters(x),
                                            }
_options_dict['HelicalParameters']['stagger'] = {
                                            'glossory': 'Stagger (Angstrom)',
                                            'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                              ' you can control the number of configurations generated in that range by increasing' +
                                                              ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                              ' Not defined for the hexad geometry.'),
                                            'default': [0.0, 0.0, 1],
                                            'validation': lambda x: _validate_helical_parameters(x),
                                            }
_options_dict['HelicalParameters']['buckle'] = {
                                            'glossory': 'Buckle (degree)',
                                            'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                              ' you can control the number of configurations generated in that range by increasing' +
                                                              ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                              ' Not defined for the hexad geometry.'),
                                            'default': [0.0, 0.0, 1],
                                            'validation': lambda x: _validate_helical_parameters(x),
                                            }
_options_dict['HelicalParameters']['propeller'] = {
                                            'glossory': 'Propeller (degree)',
                                            'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                              ' you can control the number of configurations generated in that range by increasing' +
                                                              ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                              ' Not defined for the hexad geometry.'),
                                            'default': [0.0, 0.0, 1],
                                            'validation': lambda x: _validate_helical_parameters(x),
                                            }
_options_dict['HelicalParameters']['opening'] = {
                                            'glossory': 'Opening (degree)',
                                            'long_glossory': ('Select a single value or a range of values. If you select a range of values,' +
                                                              ' you can control the number of configurations generated in that range by increasing' +
                                                              ' the number of steps. Uniformly spaced configurations in the provided range will be generated.' +
                                                              ' Not defined for the hexad geometry.'),
                                            'default': [0.0, 0.0, 1],
                                            'validation': lambda x: _validate_helical_parameters(x),
                                            }

_options_dict['HelicalParameters']['is_helical'] = {
                                                   'glossory': 'Are helical parameters used?',
                                                   'long_glossory': ("Use either helical or step parameters. The base pair parameters are " + 
                                                                     "identical in both cases"),
                                                   'default': True,
                                                   'validation': lambda x: bool(x),
                                                   }
# Runtime Parameters
_options_dict['RuntimeParameters'] = {}
_options_dict['RuntimeParameters']['search_algorithm'] = {
                                                         'glossory': 'Search algorithm',
                                                         'long_glossory': ('There are six search algorithms comprising four classes:\n' +
                                                                           '1) Monte Carlo search, 2) Random Search, 3) Genetic Algorithm Search, and 4) Systematic Search\n'
                                                                           'The first three algorithms are not deterministic. Weighted algorithms do not use' + 
                                                                           ' uniform distributions for generating random dihedral angles. Instead, the probability' +
                                                                           ' distribution for each dihedral angle is weighted by exp(-Ei/kT)/sum(exp(-Ei/kT) where' +
                                                                           ' Ei is the torsional energy at dihedral angle i.'
                                                                          ),
                                                         'default': 'systematic search',
                                                         'validation': lambda x: x.lower(),
                                                         }
_options_dict['RuntimeParameters']['num_steps'] = {
                                                  'glossory': 'Number of iterations to search over dihedral angles',
                                                  'long_glossory': ('Number of iterations (or generations in genetic algorithm) for' +
                                                                    ' searching over dihedral angles. This should be chosen such that it samples' +
                                                                    ' the dihedral angles enough while keeping the time of the search reasonbly short.' +
                                                                    ' This should be increased for systems with greater number of rotatable bonds in the backbone.'),
                                                  'default': 1000000,
                                                  'validation': lambda x: int(x),
                                                  }
_options_dict['RuntimeParameters']['seed'] = {
                                             'glossory': 'The seed for the random number generator',
                                             'long_glossory': 'Use the same value for the seed to get reproducible results in the same computer platform.',
                                             'default': 0,
                                             'validation': lambda x: int(abs(x)),
                                             }
_options_dict['RuntimeParameters']['dihedral_step'] = {
                                                      'glossory': 'Dihedral step size for systematic search (degree)',
                                                      'long_glossory': ('The number of search points for systematic search is given by:'
                                                                        ' (360/step)^(number of rotatable bonds).'),
                                                      'default': 2,
                                                      'validation': lambda x: float(x),
                                                      }
_options_dict['RuntimeParameters']['weighting_temperature'] = {
                                                              'glossory': 'Weighting temperature (K)',
                                                              'long_glossory': ('Temperature used for weighting the probability of each dihedral angle.'),
                                                              'default': 300.0,
                                                              'validation': lambda x: float(x),
                                                              }
_options_dict['RuntimeParameters']['monte_carlo_temperature'] = {
                                                                'glossory': 'Temperature used in the Monte Carlo procedure (K)',
                                                                'long_glossory': ('This temperature controls the acceptance and rejection ratio of the Monte Carlo steps.'),
                                                                'default': 300.0,
                                                                'validation': lambda x: float(x),
                                                                }
_options_dict['RuntimeParameters']['population_size'] = {
                                                        'glossory': 'The size of the population',
                                                        'long_glossory': ('The size of the population in the genetic algorithm search.'),
                                                        'default': 10000,
                                                        'validation': lambda x: int(x),
                                                        }
_options_dict['RuntimeParameters']['mutation_rate'] = {
                                                      'glossory': 'Mutation rate',
                                                      'long_glossory': ('Mutation rate in the genetic algorithm search. Used to intorduce new values' +
                                                                        ' for the dihedral angles in the population.'),
                                                      'default': 0.75,
                                                      'validation': lambda x: float(x),
                                                      }
_options_dict['RuntimeParameters']['crossover_rate'] = {
                                                       'glossory': 'Crossover rate',
                                                       'long_glossory': ('Crossover or mating rate in the genetic algorithm search.' +
                                                                         ' Used to exchange dihedral angles between individuals in the population.'),
                                                       'default': 0.25,
                                                       'validation': lambda x: float(x),
                                                       }
_options_dict['RuntimeParameters']['ff_type'] = {
                                                'glossory': 'Force field type',
                                                'long_glossory': 'Force field for computing the energy of the system.', 
                                                'default': 'MMFF94',
                                                'validation': lambda x: str(x).upper(),
                                                }
_options_dict['RuntimeParameters']['max_distance'] = {
                                                     'glossory': 'The maximum distance between atom linkers in backbone (Angstrom)',
                                                     'long_glossory': ('This is the distance between the head atom in one backbone' +
                                                                       ' and the tail atom in the adjacent backbone. When the distance' +
                                                                       ' between these two atoms is below this threshold, a bond can form between' +
                                                                       ' the two backbone molecules. This should be around 0.2 Angstroms. Conformers' +
                                                                       ' that do not pass this threshold are immediately rejected without proceeding to build' +
                                                                       ' the system.'),
                                                     'default': 0.8,
                                                     'validation': lambda x: float(x),
                                                     }
_options_dict['RuntimeParameters']['energy_filter'] = {
                                                      'glossory': ('Maximum energy for newly formed bond in the backbone (kcal/mol)\n' +
                                                                   'Maximum energy for newly formed angles in the backbone (kcal/mol)\n' +
                                                                   'Maximum torsional energy for rotatable bonds (kcal/mol/nucleotide)\n' +
                                                                   'Maximum van der Waals energy (kcal/mol/nucleotide)\n' +
                                                                   'Maximum total energy (kcal/mol/nucleotide)\n'),
                                                      'long_glossory': ('This is the bond stretching energy for the newly formed bond between the first two adjacent backbone molecules.\n' +
                                                                        'This is the angle bending energy for the newly formed angles between the first two adjacent backbone molecules.\n' +
                                                                        'This is the torsional energy for all the rotatable backbone bonds.\n' + 
                                                                        'This is the total van der Waals energy of the system.\n' + 
                                                                        'This is the total energy of the system.\n'),
                                                      'default': (1, 3, 6, 80, 10000000000),
                                                      'validation': lambda x: _validate_energy_filter(x), 
                                                      }
_options_dict['RuntimeParameters']['strand'] = {
                                               'glossory': 'FASTA string for nucleotide sequence (e.g. GCAT) ',
                                               'long_glossory': ('See the Bases section for the defined bases and their one-letter abbreviation.' + 
                                                                 'The canonical nucleobases cannot be mixed with the nucleobases that form the hexad geometries' +
                                                                 ' because they have different standard frame of reference. The names are case-insensitive.'),
                                               'default': None,
                                               'validation': lambda x: _validate_strand(x),
                                               }
_options_dict['RuntimeParameters']['pair_A_U'] = {
                                                 'glossory': 'Pair A with U (Default is A-T pairing)',
                                                 'long_glossory': 'Check box if you want to pair adenine with uracil instead of thymine.',
                                                 'default': False,
                                                 'validation': lambda x: bool(eval(x.title())) if isinstance(x, str) else bool(x),
                                                 }
_options_dict['RuntimeParameters']['is_hexad'] = {
                                                 'glossory': 'Hexad strands',
                                                 'long_glossory': 'Check box if you want to apply the 60 degrees rotation for the hexad strands.',
                                                 'default': False,
                                                 'validation': lambda x: bool(eval(x.title())) if isinstance(x, str) else bool(x),
                                                 }
_options_dict['RuntimeParameters']['build_strand'] = {
                                                     'glossory': 'Build the strand',
                                                     'long_glossory': ('Check the first box only if you want to build a single strand.' + 
                                                                       ' Check the first two boxes if you want to build a duplex.' + 
                                                                       ' Check all boxes if you want to build a hexad. Canonical nucleobases' + 
                                                                       ' can only build a single strand or a duplex.'),
                                                     'default': [True, False, False, False, False, False],
                                                     'validation': lambda x: _validate_bool_list(x),
                                                     }
_options_dict['RuntimeParameters']['strand_orientation'] = {
                                                           'glossory': 'Orientation of each strand (up or down)',
                                                           'long_glossory': ('Orientation of each strand. For DNA and RNA, check the first box' +
                                                                             ' and uncheck the second box to build anti-parallel duplex. For hexads' + 
                                                                             ' Check the boxes to build any combination of parallel and anti-parallel strands.'),
                                                           'default': [True, True, True, True, True, True],
                                                           'validation': lambda x: _validate_bool_list(x),
                                                           }
_options_dict['RuntimeParameters']['glycosidic_bond_distance'] = {
                                                                 'glossory': 'The distance of the glycosidic bond or its equivalent',
                                                                 'long_glossory': ("A user-defined distance for the glycosidic bond" + 
                                                                                   " If it is zero, the van der Waals radii are used to" + 
                                                                                   " determine an appropriate distance. This distance is set for" + 
                                                                                   " first nucleotide. The distances in the other nucleotides are" + 
                                                                                   " set by the periodic condition."),
                                                                 'default': 0.0,
                                                                 'validation': lambda x: float(x),
                                                                 }
_options_dict['RuntimeParameters']['num_candidates'] = {
                                                       'glossory': 'Quit after finding the specified number of accepted candidates',
                                                       'long_glossory': ('Quit after finding the specified number of accepted candidates for a given helical configuration.' + 
                                                                         ' Can be used to save time if only a few backbone conformations are needed.' +
                                                                         ' If the specified number of candidates is not satisfied, the search will continue' + 
                                                                         ' until the requested number of steps is completed.'),
                                                       'default': 1,
                                                       'validation': lambda x: int(x),
                                                       }



from __future__ import division, absolute_import, print_function

def test_energy_filter():
    """
    test that energy filters are enforced and that the program
    generates lower energy conformers
    """
    import os
    from io import StringIO

    import numpy as np

    from pnab import bind

    os.chdir(os.path.dirname(os.path.realpath(__file__)))

    backbone = bind.Backbone()
    backbone.file_path = os.path.join('files', 'rna_bb.pdb')
    backbone.interconnects = [10, 1]
    backbone.linker = [13, 14]

    base = bind.Base()
    base.file_path = os.path.join('files', 'adenine.pdb')
    base.code = 'A'
    base.linker = [5, 11]
    base.name = 'Adenine'
    base.pair_name = 'Uracil'
    bases = [base]

    hp = bind.HelicalParameters()
    hp.h_twist = 32.39
    hp.h_rise = 2.53
    hp.inclination = 22.9
    hp.tip = 0.08
    hp.x_displacement = -4.54
    hp.y_displacement = -0.02

    rp = bind.RuntimeParameters()
    rp.search_algorithm = 'weighted monte carlo search'
    rp.num_steps = 1
    rp.ff_type = 'GAFF'
    rp.energy_filter = [1e100]*5
    rp.max_distance = 1e100
    rp.strand = ['Adenine']*5

    output1 = bind.run(rp, backbone, bases, hp, '1')
    output1 = np.genfromtxt(StringIO(output1), delimiter=',')

    rp.num_steps = 1000000
    rp.energy_filter = [output1[3], output1[4], output1[5], output1[6], output1[7]]
    rp.max_distance = 0.05

    output2 = bind.run(rp, backbone, bases, hp, '2')
    output2 = np.genfromtxt(StringIO(output2), delimiter=',')

    if output2.size == 0:
        return
    elif output2.ndim == 1:
        output2 = output2.reshape(1, len(output2))

    filter_index = [3, 4, 5, 6, 7]
    for i in filter_index:
        assert all([val < output1[i] for val in output2[:, i]])
